from django.urls import path

from . import views

app_name = "stagedoor"
urlpatterns = [
    path("login", views.login_post, name="login"),  # type: ignore
    path("login/<str:token>", views.token_login, name="token-login"),  # type: ignore
    path("logout", views.logout, name="logout"),  # type: ignore
    path("token", views.token_post, name="token-post"),  # type: ignore
    path("approval-needed", views.approval_needed, name="approval-needed"),  # type: ignore
]



import logging
from datetime import timedelta
from random import SystemRandom

from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, AnonymousUser
from django.db import models
from django.utils.timezone import now
from phonenumber_field.modelfields import PhoneNumberField

from . import settings as stagedoor_settings

logger = logging.getLogger(__name__)

# mypy: disable-error-code="var-annotated"


class Email(models.Model):
    email = models.EmailField(unique=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_user_email",
    )
    potential_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_potential_user_email",
    )

    def __str__(self) -> str:
        return f"{self.email}: {self.user} (Maybe: {self.potential_user})"


class PhoneNumber(models.Model):
    phone_number = PhoneNumberField(
        help_text="Must include international prefix - e.g. +1 555 555 55555",
        unique=True,
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_user_phone_number",
    )
    potential_user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="stagedoor_potential_user_phone_number",
    )

    def __str__(self) -> str:
        return f"{self.phone_number}: {self.user} (Maybe: {self.potential_user})"


class AuthToken(models.Model):
    token = models.CharField(max_length=200, db_index=True)
    timestamp = models.DateTimeField(auto_now_add=True, db_index=True)
    email = models.ForeignKey(Email, blank=True, null=True, on_delete=models.CASCADE)
    phone_number = models.ForeignKey(
        PhoneNumber, blank=True, null=True, on_delete=models.CASCADE
    )
    next_url = models.CharField(max_length=2000, blank=True)
    approved = models.BooleanField(default=True)

    @classmethod
    def delete_stale(cls) -> None:
        """Delete stale tokens; tokens that are more than TOKEN_DURATION seconds old"""
        cls.objects.filter(
            timestamp__lt=now() - timedelta(seconds=stagedoor_settings.TOKEN_DURATION)
        ).delete()

    def __str__(self) -> str:
        return self.timestamp.strftime("%Y-%m-%d %H:%M:%S")  # type: ignore


def generate_token_string(sms: bool = False) -> str:
    token_length = stagedoor_settings.EMAIL_TOKEN_LENGTH
    charset = "abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ123456789"

    if sms:
        charset = "123456789"
        token_length = stagedoor_settings.SMS_TOKEN_LENGTH

    return "".join([SystemRandom().choice(charset) for _ in range(token_length)])


def generate_token(
    email: str | None = None,
    phone_number: str | None = None,
    next_url: str | None = None,
    user: AbstractBaseUser | AnonymousUser | None = None,
) -> AuthToken | None:
    created = False
    object: Email | PhoneNumber | None = None
    email_object: Email | None = None
    phone_number_object: PhoneNumber | None = None
    token_string = ""
    if email:
        token_string = generate_token_string()
        email_object, created = Email.objects.get_or_create(email=email)
        object = email_object
    if phone_number:
        token_string = generate_token_string(sms=True)
        phone_number_object, created = PhoneNumber.objects.get_or_create(
            phone_number=phone_number,
        )
        object = phone_number_object
    if not object:
        logger.error("Tried to generate a token for neither email nor sms")
        return None

    token = AuthToken(
        token=token_string,
        email=email_object,
        phone_number=phone_number_object,
        next_url=next_url or "",
    )

    if (not user or not user.is_authenticated) or created:
        token.save()
        return token
    if object.user and object.user != user:
        return None
    else:
        if isinstance(user, AbstractBaseUser):
            object.potential_user = user
            object.save()
        return token



from typing import Any

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import AbstractBaseUser
from django.http import HttpRequest

from . import settings as stagedoor_settings
from .models import AuthToken, Email, generate_token_string


class StageDoorBackend(BaseBackend):
    def get_user(self, user_id: int | str) -> AbstractBaseUser | None:
        """Get a user by their primary key."""
        User = get_user_model()
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

    def get_token_object(self, token: str | int) -> AuthToken | None:
        """Get token object by token string."""
        AuthToken.delete_stale()
        return AuthToken.objects.filter(token=token).first()

    def authenticate(
        self, request: HttpRequest | None, **kwargs: Any
    ) -> AbstractBaseUser | None:
        """Authenticate a user given a token"""
        token = kwargs.get("token")
        user = None
        AuthToken.delete_stale()

        token_object: AuthToken | None = AuthToken.objects.filter(token=token).first()
        if not token_object:
            return None

        if stagedoor_settings.SINGLE_USE_LINK:
            token_object.delete()

        User = get_user_model()

        user_args = {}

        if token_object.email and token_object.email.user:  # type: ignore[attr-defined]
            user = token_object.email.user  # type: ignore[attr-defined]
        if token_object.phone_number and token_object.phone_number.user:  # type: ignore[attr-defined]
            user = token_object.phone_number.user  # type: ignore[attr-defined]

        if not user and not stagedoor_settings.DISABLE_USER_CREATION:
            if "username" in [
                field.name for field in User._meta.get_fields(include_hidden=True)
            ]:
                user_args["username"] = f"u{generate_token_string()[:8]}"
            if token_object.email and "email" in [
                field.name for field in User._meta.get_fields(include_hidden=True)
            ]:
                user_args["email"] = token_object.email.email  # type: ignore[attr-defined]
            if token_object.phone_number and "phone_number" in [
                field.name for field in User._meta.get_fields(include_hidden=True)
            ]:
                user_args["phone_number"] = token_object.phone_number.phone_number  # type: ignore[attr-defined]

            user, _ = User.objects.get_or_create(**user_args)  # type: ignore[arg-type]

        if token_object.next_url:
            user._stagedoor_next_url = token_object.next_url  # type: ignore

        return user


class EmailTokenBackend(StageDoorBackend):
    def authenticate(
        self, request: HttpRequest | None, **kwargs: Any
    ) -> AbstractBaseUser | None:
        token = kwargs.get("token")
        if not token:
            return None
        token_object = self.get_token_object(token)
        if not token_object:
            return None
        user = super().authenticate(request, token=token_object.token)
        if not user:
            return None

        email: Email | None = token_object.email  # type: ignore[assignment]
        if not email:
            # Something has gone _real_ weird, let's be safe and return None
            return None
        if email.potential_user and email.potential_user != user:  # type: ignore[attr-defined]
            # Something has gone _real_ weird, let's be safe and return None
            return None

        email.user = user  # type: ignore[attr-defined]
        email.potential_user = None  # type: ignore[attr-defined]

        User = get_user_model()
        if "email" in [
            field.name for field in User._meta.get_fields(include_hidden=True)
        ]:
            user.email = email.email  # type: ignore
        if stagedoor_settings.SINGLE_USE_LINK:
            token_object.delete()
        user.save()
        email.save()  # type: ignore[attr-defined]
        return user


class SMSTokenBackend(StageDoorBackend):
    def authenticate(
        self, request: HttpRequest | None, **kwargs: Any
    ) -> AbstractBaseUser | None:
        token = kwargs.get("token")
        if not token:
            return None
        token_object = self.get_token_object(token)
        if not token_object:
            return None
        user = super().authenticate(request, token=token_object.token)
        if not user:
            return None

        phone_number = token_object.phone_number
        if not phone_number:
            # Something has gone _real_ weird, let's be safe and return None
            return None
        if phone_number.potential_user and phone_number.potential_user != user:  # type: ignore[attr-defined]
            # Something has gone _real_ weird, let's be safe and return None
            return None

        phone_number.user = user  # type: ignore[attr-defined]
        phone_number.potential_user = None  # type: ignore[attr-defined]

        User = get_user_model()
        if "phone_number" in [
            field.name for field in User._meta.get_fields(include_hidden=True)
        ]:
            user.phone_number = phone_number.phone_number  # type: ignore
        if stagedoor_settings.SINGLE_USE_LINK:
            token_object.delete()
        user.save()
        phone_number.save()  # type: ignore[attr-defined]
        return user



# Generated by Django 5.1.7 on 2025-04-06 04:39

import django.db.models.deletion
import phonenumber_field.modelfields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Email",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("email", models.EmailField(max_length=254, unique=True)),
                (
                    "potential_user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stagedoor_potential_user_email",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stagedoor_user_email",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PhoneNumber",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "phone_number",
                    phonenumber_field.modelfields.PhoneNumberField(
                        help_text="Must include international prefix - e.g. +1 555 555 55555",  # noqa: E501
                        max_length=128,
                        region=None,
                        unique=True,
                    ),  # type: ignore
                ),
                (
                    "potential_user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stagedoor_potential_user_phone_number",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stagedoor_user_phone_number",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AuthToken",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("token", models.CharField(db_index=True, max_length=200)),
                ("timestamp", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("next_url", models.CharField(blank=True, max_length=2000)),
                (
                    "email",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="stagedoor.email",
                    ),
                ),
                (
                    "phone_number",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="stagedoor.phonenumber",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 5.2.1 on 2025-06-14 03:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("stagedoor", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="authtoken",
            name="approved",
            field=models.BooleanField(default=True),
        ),
    ]



from django.contrib import admin

from stagedoor.helpers import email_login_link, sms_login_link

from .models import AuthToken, Email, PhoneNumber


@admin.register(Email)
class EmailAdmin(admin.ModelAdmin):
    list_display = ("email", "user", "potential_user")


@admin.register(PhoneNumber)
class PhoneNumberAdmin(admin.ModelAdmin):
    list_display = ("phone_number", "user", "potential_user")


@admin.register(AuthToken)
class AuthTokenAdmin(admin.ModelAdmin):
    list_display = ["email", "phone_number", "approved", "timestamp", "next_url"]
    ordering = ["-timestamp"]
    actions = ["approve_tokens"]

    @admin.action(description="Approve selected accounts.")
    def approve_tokens(self, request, queryset):
        """Admin action to approve selected accounts."""
        approved_count = 0

        for token in queryset:
            token.approved = True
            token.save()
            if token.email:
                email_login_link(request=request, token=token)
            if token.phone_number:
                sms_login_link(request=request, token=token)
            approved_count += 1

        self.message_user(
            request, f"Successfully approved and sent {approved_count} searches."
        )
